from aiolimiter import AsyncLimiter

from a_sync._typing import *
from a_sync import primitives
from a_sync.modifiers.manager import valid_modifiers


def get_modifiers_from(thing: Union[dict, type, object]) -> ModifierKwargs:
    if isinstance(thing, dict):
        apply_class_defined_modifiers(thing)
        return ModifierKwargs({modifier: thing[modifier] for modifier in valid_modifiers if modifier in thing})  # type: ignore [misc]
    return ModifierKwargs({modifier: getattr(thing, modifier) for modifier in valid_modifiers if hasattr(thing, modifier)})  # type: ignore [misc]

def apply_class_defined_modifiers(attrs_from_metaclass: dict):
    if 'semaphore' in attrs_from_metaclass and isinstance(val := attrs_from_metaclass['semaphore'], int):
        attrs_from_metaclass['semaphore'] = primitives.ThreadsafeSemaphore(val)
    if "runs_per_minute" in attrs_from_metaclass and isinstance(val := attrs_from_metaclass['runs_per_minute'], int):
        attrs_from_metaclass['runs_per_minute'] = AsyncLimiter(val)
        



import asyncio
import time
from threading import current_thread, main_thread
from typing import Literal

import pytest

import a_sync
from a_sync import ASyncBase
from a_sync._meta import ASyncMeta, ASyncSingletonMeta
from a_sync.singleton import ASyncGenericSingleton

increment = pytest.mark.parametrize('i', range(10))

class TestClass(ASyncBase):
    def __init__(self, v: int, sync: bool):
        self.v = v
        self.sync = sync
    
    async def test_fn(self) -> int:
        if self.sync == False:
            assert main_thread() == current_thread(), f'This should be running on an event loop in the main thread.'
        elif self.sync == True:
            assert main_thread() == current_thread(), f'This should be awaited in the main thread'
        return self.v
    
    @a_sync.aka.property
    async def test_property(self) -> int:
        if self.sync == False:
            assert main_thread() == current_thread(), f'This should be running on an event loop in the main thread.'
        elif self.sync == True:
            assert main_thread() == current_thread(), f'This should be awaited in the main thread'
        return self.v * 2
    
    @a_sync.alias.cached_property
    async def test_cached_property(self) -> int:
        if self.sync == False:
            assert main_thread() == current_thread(), f'This should be running on an event loop in the main thread.'
        elif self.sync == True:
            assert main_thread() == current_thread(), f'This should be awaited in the main thread'
        await asyncio.sleep(2)
        return self.v * 3

class TestSync(ASyncBase):
    main = main_thread()
    def __init__(self, v: int, sync: bool):
        self.v = v
        self.sync = sync
    
    def test_fn(self) -> int:
        # Sync bound methods are actually async functions that are run in an executor and awaited
        if self.sync == False:
            assert main_thread() == self.main
            assert main_thread() is not current_thread(), f'This should be running in an executor, not the main thread.'
        elif self.sync == True:
            assert main_thread() is not current_thread(), f'This should be running synchronously in the main thread'
        return self.v
    
    @a_sync.aka.property
    def test_property(self) -> int:
        if self.sync == False:
            assert main_thread() is not current_thread(), f'This should be running in an executor, not the main thread.'
        elif self.sync == True:
            # Sync properties are actually async functions that are run in an executor and awaited
            assert main_thread() is not current_thread(), f'This should be running in an executor, not the main thread.'
        return self.v * 2
    
    @a_sync.alias.cached_property
    def test_cached_property(self) -> int:
        if self.sync == False:
            assert main_thread() is not current_thread(), f'This should be running in an executor, not the main thread.'
        elif self.sync == True:
            # Sync properties are actually async functions that are run in an executor and awaited
            assert main_thread() is not current_thread(), f'This should be running in an executor, not the main thread.'
        time.sleep(2)
        return self.v * 3

class TestLimiter(TestClass):
    limiter = 1
    
class TestInheritor(TestClass):
    pass

class TestMeta(TestClass, metaclass=ASyncMeta):
    pass

class TestSingleton(ASyncGenericSingleton, TestClass):
    runs_per_minute = 100
    pass

class TestSingletonMeta(TestClass, metaclass=ASyncSingletonMeta):
    semaphore = 1
    pass

class TestSemaphore(ASyncBase):
    semaphore=1  # NOTE: this is detected propely by undecorated test_fn but not the properties
    
    def __init__(self, v: int, sync: bool):
        self.v = v
        self.sync = sync
    
    # spec on class and function both working
    #@a_sync.a_sync(semaphore=1)
    async def test_fn(self) -> int:
        await asyncio.sleep(1)
        return self.v
    

    # spec on class, function, property all working
    @a_sync.aka.property#('async', semaphore=1)
    async def test_property(self) -> int:
        await asyncio.sleep(1)
        return self.v * 2
    
    # spec on class, function, property all working
    @a_sync.alias.cached_property(semaphore=50)
    async def test_cached_property(self) -> int:
        await asyncio.sleep(1)
        return self.v * 3


def _test_kwargs(fn, default: Literal['sync','async',None]):
    # force async
    assert asyncio.get_event_loop().run_until_complete(fn(sync=False)) == 2
    assert asyncio.get_event_loop().run_until_complete(fn(asynchronous=True)) == 2
    # force sync
    with pytest.raises(TypeError):
        assert asyncio.get_event_loop().run_until_complete(fn(sync=True)) == 2
    with pytest.raises(TypeError):
        assert asyncio.get_event_loop().run_until_complete(fn(asynchronous=False)) == 2
    assert fn(sync=True) == 2
    assert fn(asynchronous=False) == 2
    if default == 'sync':
        assert fn() == 2
    elif default == 'async':
        assert asyncio.get_event_loop().run_until_complete(fn()) == 2




from typing import Any, Set


class ASyncFlagException(ValueError):
    @property
    def viable_flags(self) -> Set[str]:
        from a_sync._flags import VIABLE_FLAGS
        return VIABLE_FLAGS

    def desc(self, target) -> str:
        if target == 'kwargs':
            return "flags present in 'kwargs'"
        else:
            return f'flag attributes defined on {target}'

class NoFlagsFound(ASyncFlagException):
    def __init__(self, target, kwargs_keys=None):
        err = f"There are no viable a_sync {self.desc(target)}:"
        err += f"\nViable flags: {self.viable_flags}"
        if kwargs_keys:
            err += f"\nkwargs keys: {kwargs_keys}"
        err += "\nThis is likely an issue with a custom subclass definition."
        super().__init__(err)

class TooManyFlags(ASyncFlagException):
    def __init__(self, target, present_flags):
        err = f"There are multiple a_sync {self.__get_desc(target)} and there should only be one.\n"
        err += f"Present flags: {present_flags}\n"
        err += "This is likely an issue with a custom subclass definition."
        super().__init__(err)

class InvalidFlag(ASyncFlagException):
    def __init__(self, flag: str):
        err = f"'flag' must be one of: {self.viable_flags}. You passed {flag}."
        err += "\nThis code should not be reached and likely indicates an issue with a custom subclass definition."
        super().__init__(err)

class InvalidFlagValue(ASyncFlagException):
    def __init__(self, flag: str, flag_value: Any):
        super().__init__(f"'{flag}' should be boolean. You passed {flag_value}.")



class ImproperFunctionType(ValueError):
    pass

class FunctionNotAsync(ImproperFunctionType):
    def __init__(self, fn):
        super().__init__(f"'coro_fn' must be a coroutine function defined with 'async def'. You passed {fn}.")

class KwargsUnsupportedError(ValueError):
    def __init__(self):
        super().__init__("`run_in_executor` does not support keyword arguments, pass them as positional args instead if you're able")
        
class ASyncRuntimeError(RuntimeError):
    def __init__(self, e: RuntimeError):
        super().__init__(str(e))

class SyncModeInAsyncContextError(ASyncRuntimeError):
    def __init__(self):
        from a_sync import _flags
        err = f"The event loop is already running, which means you're trying to use an ASync function synchronously from within an async context.\n"
        err += f"Check your traceback to determine which, then try calling asynchronously instead with one of the following kwargs:\n"
        err += f"{_flags.VIABLE_FLAGS}"




from typed_envs import EnvVarFactory

envs = EnvVarFactory("EZASYNC")

# We have some envs here to help you debug your custom class implementations

# If you're only interested in debugging a specific class, set this to the class name
DEBUG_CLASS_NAME = envs.create_env("DEBUG_CLASS_NAME", str, default='', verbose=False) 

# Set this to enable debug mode on all classes
DEBUG_MODE = envs.create_env("DEBUG_MODE", bool, default=DEBUG_CLASS_NAME, verbose=False)



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import functools
from inspect import isawaitable

from a_sync import _helpers
from a_sync._typing import *
from a_sync.decorator import a_sync as unbound_a_sync
from a_sync.modified import ASyncFunction
from a_sync.property import (AsyncCachedPropertyDescriptor,
                             AsyncPropertyDescriptor)

if TYPE_CHECKING:
    from a_sync.abstract import ASyncABC


def _clean_default_from_modifiers(
    coro_fn: AsyncBoundMethod[P, T],  # type: ignore [misc]
    modifiers: dict
):
    # NOTE: We set the default here manually because the default set by the user will be used later in the code to determine whether to await.
    force_await = None
    if not asyncio.iscoroutinefunction(coro_fn) and not isinstance(coro_fn, ASyncFunction):
        if 'default' not in modifiers or modifiers['default'] != 'async':
            if 'default' in modifiers and modifiers['default'] == 'sync':
                force_await = True
            modifiers['default'] = 'async'
    return modifiers, force_await

            
def _wrap_bound_method(
    coro_fn: AsyncBoundMethod[P, T],
    **modifiers: Unpack[ModifierKwargs]
) -> AsyncBoundMethod[P, T]:
    from a_sync.abstract import ASyncABC
    
    # First we unwrap the coro_fn and rewrap it so overriding flag kwargs are handled automagically.
    if isinstance(coro_fn, ASyncFunction):
        coro_fn = coro_fn.__wrapped__
    
    modifiers, _force_await = _clean_default_from_modifiers(coro_fn, modifiers)
    
    wrapped_coro_fn: AsyncBoundMethod[P, T] = ASyncFunction(coro_fn, **modifiers)  # type: ignore [arg-type, valid-type, misc]

    @functools.wraps(coro_fn)
    def bound_a_sync_wrap(self: ASyncABC, *args: P.args, **kwargs: P.kwargs) -> T:  # type: ignore [name-defined]
        if not isinstance(self, ASyncABC):
            raise RuntimeError(f"{self} must be an instance of a class that inherits from ASyncABC.")
        # This could either be a coroutine or a return value from an awaited coroutine,
        #   depending on if an overriding flag kwarg was passed into the function call.
        retval = coro = wrapped_coro_fn(self, *args, **kwargs)
        if not isawaitable(retval):
            # The coroutine was already awaited due to the use of an overriding flag kwarg.
            # We can return the value.
            return retval  # type: ignore [return-value]
        # The awaitable was not awaited, so now we need to check the flag as defined on 'self' and await if appropriate.
        return _helpers._await(coro) if self.__a_sync_should_await__(kwargs, force=_force_await) else coro  # type: ignore [call-overload, return-value]
    return bound_a_sync_wrap

class _PropertyGetter(Awaitable[T]):
    def __init__(self, coro: Awaitable[T], property: Union[AsyncPropertyDescriptor[T], AsyncCachedPropertyDescriptor[T]]):
        self._coro = coro
        self._property = property
    def __repr__(self) -> str:
        return f"<_PropertyGetter for {self._property}._get at {hex(id(self))}>"
    def __await__(self) -> T:
        return self._coro.__await__()

@overload
def _wrap_property(
    async_property: AsyncPropertyDescriptor[T],
    **modifiers: Unpack[ModifierKwargs]
) -> AsyncPropertyDescriptor[T]:...

@overload
def _wrap_property(
    async_property: AsyncCachedPropertyDescriptor[T],
    **modifiers: Unpack[ModifierKwargs]
) -> AsyncCachedPropertyDescriptor:...

def _wrap_property(
    async_property: Union[AsyncPropertyDescriptor[T], AsyncCachedPropertyDescriptor[T]],
    **modifiers: Unpack[ModifierKwargs]
) -> Tuple[Property[T], HiddenMethod[T]]:
    if not isinstance(async_property, (AsyncPropertyDescriptor, AsyncCachedPropertyDescriptor)):
        raise TypeError(f"{async_property} must be one of: AsyncPropertyDescriptor, AsyncCachedPropertyDescriptor")

    from a_sync.abstract import ASyncABC

    async_property.hidden_method_name = f"__{async_property.field_name}__"
    
    modifiers, _force_await = _clean_default_from_modifiers(async_property, modifiers)
    
    @unbound_a_sync(**modifiers)
    async def _get(instance: ASyncABC):
        return await async_property.__get__(instance, async_property)
    
    @functools.wraps(async_property)
    def a_sync_method(self: ASyncABC, **kwargs) -> T:
        if not isinstance(self, ASyncABC):
            raise RuntimeError(f"{self} must be an instance of a class that inherits from ASyncABC.")
        awaitable = _PropertyGetter(_get(self), async_property)
        return _helpers._await(awaitable) if self.__a_sync_should_await__(kwargs, force=_force_await) else awaitable
    
    @property  # type: ignore [misc]
    @functools.wraps(async_property)
    def a_sync_property(self: ASyncABC) -> T:
        coro = getattr(self, async_property.hidden_method_name)(sync=False)
        return _helpers._await(coro) if self.__a_sync_should_await__({}, force=_force_await) else coro
    
    return a_sync_property, a_sync_method



# type: ignore [var-annotated]

import asyncio
import concurrent.futures
from decimal import Decimal
from functools import partial, wraps
from inspect import isawaitable
from typing import (Any, Awaitable, Callable, Generic, List, Optional, Set,
                    TypeVar, Union, final, overload)

from typing_extensions import ParamSpec, Self, Unpack

from a_sync._typing import ModifierKwargs

T = TypeVar('T')
P = ParamSpec('P')
MaybeAwaitable = Union[T, Awaitable[T]]

def future(callable: Union[Callable[P, Awaitable[T]], Callable[P, T]] = None, **kwargs: Unpack[ModifierKwargs]) -> Callable[P, Union[T, "ASyncFuture[T]"]]:
    return _ASyncFutureWrappedFn(callable, **kwargs)

async def _gather_check_and_materialize(*things: Unpack[MaybeAwaitable[T]]) -> List[T]:
    return await asyncio.gather(*[_check_and_materialize(thing) for thing in things])

async def _check_and_materialize(thing: T) -> T:
    return await thing if isawaitable(thing) else thing
    
def _materialize(meta: "ASyncFuture[T]") -> T:
    try:
        return asyncio.get_event_loop().run_until_complete(meta)
    except RuntimeError as e:
        raise RuntimeError(f"{meta} result is not set and the event loop is running, you will need to await it first") from e

Numeric = Union[int, float, Decimal, "ASyncFuture[int]", "ASyncFuture[float]", "ASyncFuture[Decimal]"]

class ASyncFuture(concurrent.futures.Future, Awaitable[T]):
    __slots__ = "__awaitable__", "__dependencies", "__dependants", "__task"
    def __init__(self, awaitable: Awaitable[T], dependencies: List["ASyncFuture"] = []) -> None:
        self.__awaitable__ = awaitable
        self.__dependencies = dependencies
        for dependency in dependencies:
            assert isinstance(dependency, ASyncFuture)
            dependency.__dependants.append(self)
        self.__dependants: List[ASyncFuture] = []
        self.__task = None
        super().__init__()
    def __hash__(self) -> int:
        return hash(self.__awaitable__)
    def __repr__(self) -> str:
        string = f"<{self.__class__.__name__} {self._state} for {self.__awaitable__}"
        if self.cancelled():
            pass
        elif self.done():
            string += f" exception={self.exception()}" if self.exception() else f" result={super().result()}"
        return string + ">"
    def __list_dependencies(self, other) -> List["ASyncFuture"]:
        if isinstance(other, ASyncFuture):
            return [self, other]
        return [self]
    @property
    def result(self) -> Union[Callable[[], T], Any]:
        """
        If this future is not done, it will work like cf.Future.result. It will block, await the awaitable, and return the result when ready.
        If this future is done and the result has attribute `results`, will return `getattr(future_result, 'result')`
        If this future is done and the result does NOT have attribute `results`, will again work like cf.Future.result
        """
        if self.done():
            if hasattr(r := super().result(), 'result'):
                # can be property, method, whatever. should work.
                return r.result
            # the result should be callable like an asyncio.Future
            return super().result
        return lambda: _materialize(self)
    def __getattr__(self, attr: str) -> Any:
        return getattr(_materialize(self), attr)
    def __getitem__(self, key) -> Any:
        return _materialize(self)[key]
    # NOTE: broken, do not use. I think
    def __setitem__(self, key, value) -> None:
        _materialize(self)[key] = value
    # not sure what to call these
    def __contains__(self, key: Any) -> bool:
        return _materialize(ASyncFuture(self.__contains(key), dependencies=self.__list_dependencies(key)))
    def __await__(self) -> Awaitable[T]:
        return self.__await().__await__()
    async def __await(self) -> T:
        if not self.done():
            self.set_result(await self.__task__)
        return self._result
    @property
    def __task__(self) -> "asyncio.Task[T]":
        if self.__task is None:
            self.__task = asyncio.create_task(self.__awaitable__)
        return self.__task
    def __iter__(self):
        return _materialize(self).__iter__()
    def __next__(self):
        return _materialize(self).__next__()
    def __enter__(self):
        return _materialize(self).__enter__()
    def __exit__(self, *args):
        return _materialize(self).__exit__(*args)
    @overload
    def __add__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    def __add__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    def __add__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __add__(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    def __add__(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    def __add__(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    def __add__(self, other: Numeric) -> "ASyncFuture":
        return ASyncFuture(self.__add(other), dependencies=self.__list_dependencies(other))
    @overload
    def __sub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    def __sub__(self, other: Numeric) -> "ASyncFuture":
        return ASyncFuture(self.__sub(other), dependencies=self.__list_dependencies(other))
    def __mul__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__mul(other), dependencies=self.__list_dependencies(other))
    def __pow__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__pow(other), dependencies=self.__list_dependencies(other))
    def __truediv__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__truediv(other), dependencies=self.__list_dependencies(other))
    def __floordiv__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__floordiv(other), dependencies=self.__list_dependencies(other))
    def __pow__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__pow(other), dependencies=self.__list_dependencies(other))
    @overload
    def __radd__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __radd__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __radd__(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    def __radd__(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    def __radd__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__radd(other), dependencies=self.__list_dependencies(other))
    
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __rsub__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    def __rsub__(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    def __rsub__(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    def __rsub__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__rsub(other), dependencies=self.__list_dependencies(other))
    def __rmul__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__rmul(other), dependencies=self.__list_dependencies(other))
    def __rtruediv__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__rtruediv(other), dependencies=self.__list_dependencies(other))
    def __rfloordiv__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__rfloordiv(other), dependencies=self.__list_dependencies(other))
    def __rpow__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__rpow(other), dependencies=self.__list_dependencies(other))
    def __eq__(self, other) -> "ASyncFuture":
        return bool(ASyncFuture(self.__eq(other), dependencies=self.__list_dependencies(other)))
    def __gt__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__gt(other), dependencies=self.__list_dependencies(other))
    def __ge__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__ge(other), dependencies=self.__list_dependencies(other))
    def __lt__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__lt(other), dependencies=self.__list_dependencies(other))
    def __le__(self, other) -> "ASyncFuture":
        return ASyncFuture(self.__le(other), dependencies=self.__list_dependencies(other))
    
    # Maths
    
    @overload
    async def __add(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    async def __add(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __add(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    async def __add(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __add(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    async def __add(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __add(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __add(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __add(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    async def __add(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a + b
    @overload
    async def __sub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    async def __sub(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __sub(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    async def __sub(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __sub(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __sub(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    async def __sub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a - b
    async def __mul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a * b
    async def __truediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a / b
    async def __floordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a // b
    async def __pow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a ** b
    
    # rMaths
    @overload
    async def __radd(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    async def __radd(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __radd(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    async def __radd(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __radd(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    async def __radd(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __radd(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __radd(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __radd(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    async def __radd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a + b
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]":...
    @overload
    async def __rsub(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[Decimal]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __rsub(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]":...
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]":...
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: Awaitable[int]) -> "ASyncFuture[int]":...
    @overload
    async def __rsub(self: "ASyncFuture[float]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[float]", other: Awaitable[int]) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: Awaitable[float]) -> "ASyncFuture[float]":...
    @overload
    async def __rsub(self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __rsub(self: "ASyncFuture[Decimal]", other: Awaitable[int]) -> "ASyncFuture[Decimal]":...
    @overload
    async def __rsub(self: "ASyncFuture[int]", other: Awaitable[Decimal]) -> "ASyncFuture[Decimal]":...
    async def __rsub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a - b
    async def __rmul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a * b
    async def __rtruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a / b
    async def __rfloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a // b
    async def __rpow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a ** b
    
    async def __iadd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a + b
        return self._result
    async def __isub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a - b
        return self._result
    async def __imul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a * b
        return self._result
    async def __itruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a / b
        return self._result
    async def __ifloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a // b
        return self._result
    async def __ipow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a ** b
        return self._result
    
    # Comparisons
    async def __eq(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a == b
    async def __gt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a > b
    async def __ge(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a >= b
    async def __lt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a < b
    async def __le(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a <= b

    
    # not sure what to call these
    async def __contains(self, item: Any) -> bool:
        _self, _item = await _gather_check_and_materialize(self, item)
        return _item in _self
        
    # conversion
    # NOTE: We aren't allowed to return ASyncFutures here :(
    def __bool__(self) -> bool:
        return bool(_materialize(self))
    def __bytes__(self) -> bytes:
        return bytes(_materialize(self))
    def __str__(self) -> str:
        return str(_materialize(self))
    def __int__(self) -> int:
        return int(_materialize(self))
    def __float__(self) -> float:
        return float(_materialize(self))
    
    # WIP internals
    
    @property
    def __dependants__(self) -> Set["ASyncFuture"]:
        dependants = set()
        for dep in self.__dependants:
            dependants.add(dep)
            dependants.union(dep.__dependants__)
        return dependants
    @property
    def __dependencies__(self) -> Set["ASyncFuture"]:
        dependencies = set()
        for dep in self.__dependencies:
            dependencies.add(dep)
            dependencies.union(dep.__dependencies__)
        return dependencies
    def __sizeof__(self) -> int:
        if isinstance(self.__awaitable__, Coroutine):
            return sum(sys.getsizeof(v) for v in self.__awaitable__.cr_frame.f_locals.values())
        elif isinstance(self.__awaitable__, asyncio.Future):
            raise NotImplementedError
        raise NotImplementedError


@final  
class _ASyncFutureWrappedFn(Callable[P, ASyncFuture[T]]):
    __slots__ = "callable", "wrapped", "_callable_name"
    def __init__(self, callable: Union[Callable[P, Awaitable[T]], Callable[P, T]] = None, **kwargs: Unpack[ModifierKwargs]):
        from a_sync import a_sync
        if callable:
            self.callable = callable
            self._callable_name = callable.__name__
            a_sync_callable = a_sync(callable, default="async", **kwargs)
            @wraps(callable)
            def future_wrap(*args: P.args, **kwargs: P.kwargs) -> "ASyncFuture[T]":
                return ASyncFuture(a_sync_callable(*args, **kwargs, sync=False))
            self.wrapped = future_wrap
        else:
            self.wrapped = partial(_ASyncFutureWrappedFn, **kwargs)
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ASyncFuture[T]:
        return self.wrapped(*args, **kwargs)
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.callable}>"
    def __get__(self, instance: object, owner: Optional) -> Union[Self, "_ASyncFutureInstanceMethod[P, T]"]:
        if owner is None:
            return self
        else:
            return _ASyncFutureInstanceMethod(self, instance)

@final
class _ASyncFutureInstanceMethod(Generic[P, T]):
    # NOTE: probably could just replace this with functools.partial
    def __init__(
        self,
        wrapper: _ASyncFutureWrappedFn[P, T],
        instance: object,
    ) -> None:
        try:
            self.__module__ = wrapper.__module__
        except AttributeError:
            pass
        try:
            self.__name__ = wrapper.__name__
        except AttributeError:
            pass
        try:
            self.__qualname__ = wrapper.__qualname__
        except AttributeError:
            pass
        try:
            self.__doc__ = wrapper.__doc__
        except AttributeError:
            pass
        try:
            self.__annotations__ = wrapper.__annotations__
        except AttributeError:
            pass
        try:
            self.__dict__.update(wrapper.__dict__)
        except AttributeError:
            pass
        self.__instance = instance
        self.__wrapper = wrapper

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} for {self.__wrapper.callable} bound to {self.__instance}>"
    def __call__(self, /, *fn_args: P.args, **fn_kwargs: P.kwargs) -> T:
        return self.__wrapper(self.__instance, *fn_args, **fn_kwargs)




import asyncio
import logging
from asyncio.futures import _chain_future
from typing import AsyncIterator, Optional, TypeVar, Union, overload

from a_sync.primitives.queue import Queue

logger = logging.getLogger(__name__)

T = TypeVar('T')

async def exhaust_iterator(iterator: AsyncIterator[T], *, queue: Optional[asyncio.Queue] = None) -> None:
    async for thing in iterator:
        if queue:
            logger.debug('putting %s from %s to queue %s', thing, iterator, queue)
            queue.put_nowait(thing)
        
async def exhaust_iterators(iterators, *, queue: Optional[asyncio.Queue] = None) -> None:
    await asyncio.gather(*[exhaust_iterator(iterator, queue=queue) for iterator in iterators]) 
    if queue:
        queue.put_nowait(_Done())
    
T0 = TypeVar('T0')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')
T7 = TypeVar('T7')
T8 = TypeVar('T8')
T9 = TypeVar('T9')

@overload
async def as_yielded(*iterators: AsyncIterator[T]) -> AsyncIterator[T]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7], iterator8: AsyncIterator[T8], iterator9: AsyncIterator[T9]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7, T8, T9]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7], iterator8: AsyncIterator[T8]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7, T8]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4]) -> AsyncIterator[Union[T0, T1, T2, T3, T4]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3]) -> AsyncIterator[Union[T0, T1, T2, T3]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2]) -> AsyncIterator[Union[T0, T1, T2]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1]) -> AsyncIterator[Union[T0, T1]]:...
@overload
async def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], *iterators: AsyncIterator[T]) -> AsyncIterator[Union[T0, T1, T2, T]]:...
async def as_yielded(*iterators: AsyncIterator[T]) -> AsyncIterator[T]:  # type: ignore [misc]
    queue = Queue()
    task = asyncio.create_task(exhaust_iterators(iterators, queue=queue))
    def done_callback(t: asyncio.Task) -> None:
        if t.exception() and not next_fut.done(): 
            next_fut.set_exception(t.exception())
        
    task.add_done_callback(done_callback)
    while not task.done():
        next_fut = asyncio.get_event_loop().create_future()
        get_task = asyncio.create_task(coro=queue.get(), name=str(queue))
        _chain_future(get_task, next_fut)
        for item in (await next_fut, *queue.get_nowait(-1)):
            if isinstance(item, _Done):
                task.cancel()
                return
            yield item
            
    if e := task.exception():
        get_task.cancel()
        raise e

class _Done:
    pass



import asyncio
import heapq
import logging
from collections import deque
from functools import cached_property
from typing import (Deque, Dict, Generic, List, Literal, Optional, Protocol,
                    Type, TypeVar)

from a_sync.primitives.locks.semaphore import Semaphore

logger = logging.getLogger(__name__)

T = TypeVar('T', covariant=True)

class Priority(Protocol):
    def __lt__(self, other) -> bool:
        ...

PT = TypeVar('PT', bound=Priority)
    
CM = TypeVar('CM', bound="_AbstractPrioritySemaphoreContextManager[Priority]")

class _AbstractPrioritySemaphore(Semaphore, Generic[PT, CM]):
    name: Optional[str]
    _value: int
    _waiters: List["_AbstractPrioritySemaphoreContextManager[PT]"]  # type: ignore [assignment]

    @property
    def _context_manager_class(self) -> Type["_AbstractPrioritySemaphoreContextManager[PT]"]:
        raise NotImplementedError
    
    @property
    def _top_priority(self) -> PT:
        # You can use this so you can set priorities with non numeric comparable values
        raise NotImplementedError

    def __init__(self, value: int = 1, *, name: Optional[str] = None) -> None:
        self._context_managers: Dict[PT, _AbstractPrioritySemaphoreContextManager[PT]] = {}
        self._capacity = value
        super().__init__(value, name=name)
        self._waiters = []
        # NOTE: This should (hopefully) be temporary
        self._potential_lost_waiters = []

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name} capacity={self._capacity} value={self._value} waiters={self._count_waiters()}>"

    async def __aenter__(self) -> None:
        await self[self._top_priority].acquire()

    async def __aexit__(self, *_) -> None:
        self[self._top_priority].release()
    
    async def acquire(self) -> Literal[True]:
        return await self[self._top_priority].acquire()
    
    def __getitem__(self, priority: Optional[PT]) -> "_AbstractPrioritySemaphoreContextManager[PT]":
        priority = self._top_priority if priority is None else priority
        if priority not in self._context_managers:
            context_manager = self._context_manager_class(self, priority, name=self.name)
            heapq.heappush(self._waiters, context_manager)  # type: ignore [misc]
            self._context_managers[priority] = context_manager
        return self._context_managers[priority]

    def locked(self) -> bool:
        """Returns True if semaphore cannot be acquired immediately."""
        return self._value == 0 or (
            any(
                cm._waiters and any(not w.cancelled() for w in cm._waiters) 
                for cm in (self._context_managers.values() or ())
            )
        )
    
    def _count_waiters(self) -> Dict[PT, int]:
        return {manager._priority: len(manager.waiters) for manager in sorted(self._waiters, key=lambda m: m._priority)}
    
    def _wake_up_next(self) -> None:
        while self._waiters:
            manager = heapq.heappop(self._waiters)
            if len(manager) == 0:
                # There are no more waiters, get rid of the empty manager
                logger.debug("manager %s has no more waiters, popping from %s", manager._repr_no_parent_(), self)
                self._context_managers.pop(manager._priority)
                continue
            logger.debug("waking up next for %s", manager._repr_no_parent_())
            
            woke_up = False
            start_len = len(manager)
        
            if not manager._waiters:
                logger.debug('not manager._waiters')
            
            while manager._waiters:
                waiter = manager._waiters.popleft()
                self._potential_lost_waiters.remove(waiter)
                if not waiter.done():
                    waiter.set_result(None)
                    logger.debug("woke up %s", waiter)
                    woke_up = True
                    break
            
            if not woke_up:
                self._context_managers.pop(manager._priority)
                continue
            
            end_len = len(manager)
            
            assert start_len > end_len, f"start {start_len} end {end_len}"
            
            if end_len:
                # There are still waiters, put the manager back
                heapq.heappush(self._waiters, manager)  # type: ignore [misc]
            else:
                # There are no more waiters, get rid of the empty manager
                self._context_managers.pop(manager._priority)
            return
        
        # emergency procedure (hopefully temporary): 
        while self._potential_lost_waiters:
            waiter = self._potential_lost_waiters.pop(0)
            logger.debug('we found a lost waiter %s', waiter)
            if not waiter.done():
                waiter.set_result(None)
                logger.debug("woke up lost waiter %s", waiter)
                return
        logger.debug("%s has no waiters to wake", self)

class _AbstractPrioritySemaphoreContextManager(Semaphore, Generic[PT]):
    _loop: asyncio.AbstractEventLoop
    _waiters: Deque[asyncio.Future]  # type: ignore [assignment]
    
    @property
    def _priority_name(self) -> str:
        raise NotImplementedError
    
    def __init__(self, parent: _AbstractPrioritySemaphore, priority: PT, name: Optional[str] = None) -> None:
        self._parent = parent
        self._priority = priority
        super().__init__(0, name=name)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} parent={self._parent} {self._priority_name}={self._priority} waiters={len(self)}>"
    
    def _repr_no_parent_(self) -> str:
        return f"<{self.__class__.__name__} parent_name={self._parent.name} {self._priority_name}={self._priority} waiters={len(self)}>"
    
    def __lt__(self, other) -> bool:
        if type(other) is not type(self):
            raise TypeError(f"{other} is not type {self.__class__.__name__}")
        return self._priority < other._priority
    
    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop or asyncio.get_event_loop()
    
    @property
    def waiters (self) -> Deque[asyncio.Future]:
        if self._waiters is None:
            self._waiters = deque()
        return self._waiters
    
    async def acquire(self) -> Literal[True]:
        """Acquire a semaphore.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately.  If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.
        """
        if self._parent._value <= 0:
            self._ensure_debug_daemon()
        while self._parent._value <= 0:
            fut = self.loop.create_future()
            self.waiters.append(fut)
            self._parent._potential_lost_waiters.append(fut)
            try:
                await fut
            except:
                # See the similar code in Queue.get.
                fut.cancel()
                if self._parent._value > 0 and not fut.cancelled():
                    self._parent._wake_up_next()
                raise
        self._parent._value -= 1
        return True
    def release(self) -> None:
        self._parent.release()
    
class _PrioritySemaphoreContextManager(_AbstractPrioritySemaphoreContextManager[int]):
    _priority_name = "priority"

class PrioritySemaphore(_AbstractPrioritySemaphore[int, _PrioritySemaphoreContextManager]):  # type: ignore [type-var]
    _context_manager_class = _PrioritySemaphoreContextManager
    _top_priority = -1
    """
    It's kinda like a regular Semaphore but you must give each waiter a priority:

    ```
    priority_semaphore = PrioritySemaphore(10)

    async with priority_semaphore[priority]:
        await do_stuff()
    ```
    
    You can aenter and aexit this semaphore without a priority and it will process those first. Like so:
    
    ```
    priority_semaphore = PrioritySemaphore(10)
    
    async with priority_semaphore:
        await do_stuff()
    ```
    """



import asyncio
import sys
from typing import Generic, List, TypeVar, Union, overload

T = TypeVar('T')

if sys.version_info < (3, 9):
    bases = (asyncio.Queue, Generic[T])
else:
    bases = (asyncio.Queue[T], )

class Queue(*bases):
    """The only difference between an a_sync.Queue and an asyncio.Queue is that `get_nowait` can retrn multiple responses."""
    @overload
    def get_nowait(self, i = 1, can_return_less: bool = False) -> T:
        ...
    @overload
    def get_nowait(self, i: int, can_return_less: bool = False) -> List[T]:
        ...
    def get_nowait(self, i: int = 1, can_return_less: bool = False) -> Union[T, List[T]]:
        """
        Just like `asyncio.Queue.get_nowait`, but will return `i` items instead of 1.
        Set `can_return_less` to True if you want to receive up to `i` items.
        """
        if can_return_less and i == 1:
            raise ValueError("you cant set i == 1 with can_return_less == True")
        values = []
        if i == -1:
            while True:
                try:
                    values.append(super().get_nowait())
                except asyncio.QueueEmpty:
                    return values
        for _ in range(i):
            try:
                values.append(super().get_nowait())
            except asyncio.QueueEmpty:
                if can_return_less:
                    break
                raise
        return values[0] if i == 1 else values



from a_sync.primitives.locks.counter import CounterLock
from a_sync.primitives.locks.event import Event
from a_sync.primitives.locks.semaphore import DummySemaphore, Semaphore, ThreadsafeSemaphore
from a_sync.primitives.locks.prio_semaphore import PrioritySemaphore




"""
While not the focus of this lib, this module includes some new primitives and some modified versions of standard asyncio primitives.
"""

from a_sync.primitives.executor import (AsyncProcessPoolExecutor,
                                        AsyncThreadPoolExecutor,
                                        ProcessPoolExecutor,
                                        PruningThreadPoolExecutor,
                                        ThreadPoolExecutor)
from a_sync.primitives.locks import *
from a_sync.primitives.queue import Queue



import asyncio
import functools
import logging
from collections import defaultdict
from threading import Thread, current_thread
from typing import Callable, Literal, Optional, TypeVar

from typing_extensions import ParamSpec

from a_sync.primitives._debug import _DebugDaemonMixin

logger = logging.getLogger(__name__)

T = TypeVar('T')
P = ParamSpec('P')

class Semaphore(asyncio.Semaphore, _DebugDaemonMixin):
    def __init__(self, value: int, name=None, **kwargs) -> None:
        """
        `name` is used only in debug logs to provide useful context
        """
        super().__init__(value, **kwargs)
        self.name = name or self.__origin__ if hasattr(self, '__origin__') else None
        self._decorated = set()
    
    # Dank new functionality
    def __call__(self, fn: Callable[P, T]) -> Callable[P, T]:
        return self.decorate(fn)
    
    def __repr__(self) -> str:
        representation = f"<{self.__class__.__name__} name={self.name} value={self._value} waiters={len(self)}>"
        if self._decorated:
            representation = f"{representation[:-1]} decorates={self._decorated}"
        return representation
    
    def __len__(self) -> int:
        return len(self._waiters) if self._waiters else 0
    
    def decorate(self, fn: Callable[P, T]) -> Callable[P, T]:
        if not asyncio.iscoroutinefunction(fn):
            raise TypeError(f"{fn} must be a coroutine function")
        @functools.wraps(fn)
        async def semaphore_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with self:
                return await fn(*args, **kwargs)
        self._decorated.add(f"{fn.__module__}.{fn.__name__}")
        return semaphore_wrapper
    
    async def acquire(self) -> bool:
        if self._value <= 0:
            self._ensure_debug_daemon()
        return await super().acquire()
    
    # Everything below just adds some debug logs
    async def _debug_daemon(self) -> None:
        while self._waiters:
            await asyncio.sleep(60)
            self.logger.debug(f"{self} has {len(self)} waiters for any of: {self._decorated}")
  
        
class DummySemaphore(asyncio.Semaphore):
    """It can go where a semaphore goes, but it does nothing."""
    _value = 0
    def __init__(self, name: Optional[str] = None):
        self.name = name
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name}>"
    async def acquire(self) -> Literal[True]:
        return True
    def release(self) -> None:
        ...
    async def __aenter__(self):
        ...
    async def __aexit__(self, *args):
        ...
        

class ThreadsafeSemaphore(Semaphore):
    """
    While its a bit weird to run multiple event loops, sometimes either you or a lib you're using must do so. 
    When in use in threaded applications, this semaphore will not work as intended but at least your program will function.
    You may need to reduce the semaphore value for multi-threaded applications.
    
    # TL;DR it's a janky fix for an edge case problem and will otherwise function as a normal a_sync.Semaphore (which is just an asyncio.Semaphore with extra bells and whistles).
    """

    def __init__(self, value: Optional[int], name: Optional[str] = None) -> None:
        assert isinstance(value, int), f"{value} should be an integer."
        super().__init__(value, name=name)
        self.semaphores: DefaultDict[Thread, Semaphore] = defaultdict(lambda: Semaphore(value, name=self.name))  # type: ignore [arg-type]
        self.dummy = DummySemaphore(name=name)
    
    def __len__(self) -> int:
        return sum(len(sem._waiters) for sem in self.semaphores.values())
    
    @property
    def use_dummy(self) -> bool:
        return self._value is None
    
    @property
    def semaphore(self) -> Semaphore:
        return self.dummy if self.use_dummy else self.semaphores[current_thread()]
    
    async def __aenter__(self):
        await self.semaphore.acquire()
    
    async def __aexit__(self, *args):
        self.semaphore.release()
        



import asyncio
import time

import pytest

from a_sync._meta import ASyncMeta
from tests.fixtures import TestClass, TestInheritor, TestMeta, increment, TestSync

classes = pytest.mark.parametrize('cls', [TestClass, TestSync, TestInheritor, TestMeta])

@classes
@increment
def test_base_sync(cls: type, i: int):
    sync_instance = cls(i, True)
    assert isinstance(sync_instance.__class__, ASyncMeta)

    assert sync_instance.test_fn() == i
    assert sync_instance.test_property == i * 2
    start = time.time()
    assert sync_instance.test_cached_property == i * 3
    assert isinstance(sync_instance.test_cached_property, int)
    duration = time.time() - start
    assert duration < 3, "There is a 2 second sleep in 'test_cached_property' but it should only run once."

    # Can we override with kwargs?
    val = asyncio.get_event_loop().run_until_complete(sync_instance.test_fn(sync=False))
    assert isinstance(val, int)

    # Can we access hidden methods for properties?
    assert sync_instance.__test_property__() == i * 2
    start = time.time()
    assert sync_instance.__test_cached_property__() == i * 3
    # Can we override them too?
    assert asyncio.get_event_loop().run_until_complete(sync_instance.__test_cached_property__(sync=False)) == i * 3
    duration = time.time() - start
    assert duration < 3, "There is a 2 second sleep in 'test_cached_property' but it should only run once."

@classes
@increment
@pytest.mark.asyncio_cooperative
async def test_base_async(cls: type, i: int):
    async_instance = cls(i, False)
    assert isinstance(async_instance.__class__, ASyncMeta)
    
    assert await async_instance.test_fn() == i
    assert await async_instance.test_property == i * 2
    start = time.time()
    assert await async_instance.test_cached_property == i * 3
    assert isinstance(await async_instance.test_cached_property, int)
    duration = time.time() - start
    target_duration = 5 if isinstance(async_instance, TestSync) else 3
    # For TestSync, the duration can be higher because the calls execute inside of a threadpool which limits the amount of concurrency.
    assert duration < target_duration, "There is a 2 second sleep in 'test_cached_property' but it should only run once."

    #if isinstance(async_instance, TestSync):
    #    # NOTE this shoudl probbaly run sync in main thread instead of raising...
    #    with pytest.raises(RuntimeError):
    #        await async_instance.test_fn()
        
    # Can we override with kwargs?
    
    if not isinstance(async_instance, TestSync):
        with pytest.raises(RuntimeError):
            async_instance.test_fn(sync=True)
    
    # Can we access hidden methods for properties?
    assert await async_instance.__test_property__() == i * 2
    assert await async_instance.__test_cached_property__() == i * 3
    # Can we override them too?
    with pytest.raises(RuntimeError):
        async_instance.__test_cached_property__(sync=True)




"""
With these executors, you can simply run sync fns in your executor with `await executor.run(fn, *args)`

`executor.submit(fn, *args)` will work the same as the concurrent.futures implementation, but will return an asyncio.Future instead of a concurrent.futures.Future
"""

import asyncio
import concurrent.futures as cf
import multiprocessing.context
import queue
import threading
import weakref
from concurrent.futures import _base, thread
from functools import cached_property
from typing import Any, Callable, Optional, Tuple, TypeVar

from typing_extensions import ParamSpec

from a_sync.primitives._debug import _DebugDaemonMixin

TEN_MINUTES = 60 * 10

T = TypeVar('T')
P = ParamSpec('P')


class _AsyncExecutorMixin(cf.Executor, _DebugDaemonMixin):
    _max_workers: int
    _workers: str
    async def run(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        """
        A shorthand way to call `await asyncio.get_event_loop().run_in_executor(this_executor, fn, *args)`
        Doesn't `await this_executor.run(fn, *args)` look so much better?
        
        Oh, and you can also use kwargs!
        """
        return fn(*args, **kwargs) if self.sync_mode else await self.submit(fn, *args, **kwargs)
    def submit(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[T]":
        """Submits a job to the executor and returns an `asyncio.Future` that can be awaited for the result without blocking."""
        if self.sync_mode:
            fut = asyncio.ensure_future(self._exec_sync(fn, *args, **kwargs))
        else:
            fut = asyncio.futures.wrap_future(super().submit(fn, *args, **kwargs))
            self._start_debug_daemon(fut, fn, *args, **kwargs)
        return fut
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} object at {hex(id(self))} [{self.worker_count_current}/{self._max_workers} {self._workers}]>"
    def __len__(self) -> int:
        # NOTE: should this be queue length instead? probably
        return self.worker_count_current
    @cached_property
    def sync_mode(self) -> bool:
        return self._max_workers == 0
    @property
    def worker_count_current(self) -> int:
        len(getattr(self, f"_{self._workers}"))
    async def _exec_sync(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        """Just wraps a fn and its args into an awaitable."""
        return fn(*args, **kwargs)
    async def _debug_daemon(self, fut: asyncio.Future, fn, *args, **kwargs) -> None:
        """Runs until manually cancelled by the finished work item"""
        while not fut.done():
            await asyncio.sleep(15)
            if not fut.done():
                self.logger.debug(f'{self} processing {fn}{args}{kwargs}')
    
# Process

class AsyncProcessPoolExecutor(_AsyncExecutorMixin, cf.ProcessPoolExecutor):
    _workers = "processes"
    def __init__(
        self, 
        max_workers: Optional[int] = None, 
        mp_context: Optional[multiprocessing.context.BaseContext] = None, 
        initializer: Callable[..., object] = None,
        initargs: Tuple[Any, ...] = (),
    ) -> None:
        if max_workers == 0:
            super().__init__(1, mp_context, initializer, initargs)
            self._max_workers = 0
        else:
            super().__init__(max_workers, mp_context, initializer, initargs)

# Thread

class AsyncThreadPoolExecutor(_AsyncExecutorMixin, cf.ThreadPoolExecutor):
    _workers = "threads"
    def __init__(
        self, 
        max_workers: Optional[int] = None, 
        thread_name_prefix: str = '', 
        initializer: Callable[..., object] = None,
        initargs: Tuple[Any, ...] = (),
    ) -> None:
        if max_workers == 0:
            super().__init__(1, thread_name_prefix, initializer, initargs)
            self._max_workers = 0
        else:
            super().__init__(max_workers, thread_name_prefix, initializer, initargs)
    
# For backward-compatibility
ProcessPoolExecutor = AsyncProcessPoolExecutor
ThreadPoolExecutor = AsyncThreadPoolExecutor

# Pruning thread pool

def _worker(executor_reference, work_queue, initializer, initargs, timeout):  # NOTE: NEW 'timeout'
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            _base.LOGGER.critical('Exception in initializer:', exc_info=True)
            executor = executor_reference()
            if executor is not None:
                executor._initializer_failed()
            return
        
    try:
        while True:
            try:  # NOTE: NEW
                work_item = work_queue.get(block=True, 
                                           timeout=timeout)  # NOTE: NEW
            except queue.Empty:  # NOTE: NEW
                # Its been 'timeout' seconds and there are no new work items.  # NOTE: NEW
                # Let's suicide the thread.  # NOTE: NEW
                executor = executor_reference()  # NOTE: NEW
                
                with executor._adjusting_lock:  # NOTE: NEW
                    # NOTE: We keep a minimum of one thread active to prevent locks
                    if len(executor) > 1:  # NOTE: NEW
                        t = threading.current_thread()  # NOTE: NEW
                        executor._threads.remove(t)  # NOTE: NEW
                        thread._threads_queues.pop(t)  # NOTE: NEW
                        # Let the executor know we have one less idle thread available
                        executor._idle_semaphore.acquire(blocking=False)  # NOTE: NEW
                        return  # NOTE: NEW 
                continue
                
            if work_item is not None:
                work_item.run()
                # Delete references to object. See issue16284
                del work_item

                # attempt to increment idle count
                executor = executor_reference()
                if executor is not None:
                    executor._idle_semaphore.release()
                del executor
                continue

            executor = executor_reference()
            # Exit if:
            #   - The interpreter is shutting down OR
            #   - The executor that owns the worker has been collected OR
            #   - The executor that owns the worker has been shutdown OR
            if thread._shutdown or executor is None or executor._shutdown:
                # Flag the executor as shutting down as early as possible if it
                # is not gc-ed yet.
                if executor is not None:
                    executor._shutdown = True
                # Notice other workers
                work_queue.put(None)
                return
            del executor
    except BaseException:
        _base.LOGGER.critical('Exception in worker', exc_info=True)

class PruningThreadPoolExecutor(AsyncThreadPoolExecutor):
    """
    This ThreadPoolExecutor implementation prunes inactive threads after 'timeout' seconds without a work item.
    Pruned threads will be automatically recreated as needed for future workloads. up to 'max_threads' can be active at any one time.
    """
    def __init__(self, max_workers=None, thread_name_prefix='',
                 initializer=None, initargs=(), timeout=TEN_MINUTES):
        self._timeout=timeout
        self._adjusting_lock = threading.Lock()
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
    
    def __len__(self) -> int:
        return len(self._threads)
        
    def _adjust_thread_count(self):
        with self._adjusting_lock:
            # if idle threads are available, don't spin new threads
            if self._idle_semaphore.acquire(timeout=0):
                return

            # When the executor gets lost, the weakref callback will wake up
            # the worker threads.
            def weakref_cb(_, q=self._work_queue):
                q.put(None)

            num_threads = len(self._threads)
            if num_threads < self._max_workers:
                thread_name = '%s_%d' % (self._thread_name_prefix or self,
                                         num_threads)
                t = threading.Thread(name=thread_name, target=_worker,
                                     args=(weakref.ref(self, weakref_cb),
                                           self._work_queue,
                                           self._initializer,
                                           self._initargs,
                                           self._timeout))
                t.daemon = True
                t.start()
                self._threads.add(t)
                thread._threads_queues[t] = self._work_queue

executor = PruningThreadPoolExecutor(128)




import asyncio
from typing import (Any, Awaitable, Dict, List, Mapping, TypeVar, Union,
                    overload)

try:
    from tqdm.asyncio import tqdm_asyncio
except ImportError as e:
    class tqdm_asyncio:  # type: ignore [no-redef]
        async def gather(*args, **kwargs):
            raise ImportError("You must have tqdm installed in order to use this feature")

from a_sync.utils.as_completed import as_completed_mapping

T = TypeVar('T')
KT = TypeVar('KT')
VT = TypeVar('VT')

@overload
async def gather(*awaitables: Mapping[KT, Awaitable[VT]], return_exceptions: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Dict[KT, VT]:
    ...
@overload
async def gather(*awaitables: Awaitable[T], return_exceptions: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> List[T]:
    ...
async def gather(*awaitables: Union[Awaitable[T], Mapping[KT, Awaitable[VT]]], return_exceptions: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Union[List[T], Dict[KT, VT]]:
    """
    Concurrently awaits a list of awaitable objects or mappings of awaitables and returns the results.

    This function extends Python's asyncio.gather, providing additional features for mixed use cases of individual awaitable objects and mappings of awaitables.

    Differences from asyncio.gather:
    - Uses type hints for use with static type checkers.
    - Supports gathering either individual awaitables or a k:v mapping of awaitables.
    - Provides progress reporting using tqdm if 'tqdm' is set to True.
    
    Args:
        *awaitables (Union[Awaitable[T], Mapping[KT, Awaitable[VT]]]): The awaitables to await concurrently. It can be a single awaitable or a mapping of awaitables.
        return_exceptions (bool, optional): If True, exceptions are returned as results instead of raising them. Defaults to False.
        tqdm (bool, optional): If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Returns:
        Union[List[T], Dict[KT, VT]]: A list of results when awaiting individual awaitables or a dictionary of results when awaiting mappings.

    Examples:
        Awaiting individual awaitables:
        results will be a list containing the result of each awaitable in sequential order
        ```
        results = await gather(thing1(), thing2())
        ```

        Awaiting mappings of awaitables
        results will be a dictionary with 'key1' mapped to the result of thing1() and 'key2' mapped to the result of thing2.
        ```
        mapping = {'key1': thing1(), 'key2': thing2()}
        results = await gather(mapping)
        ```
    """
    return await (
        gather_mapping(awaitables[0], return_exceptions=return_exceptions, tqdm=tqdm, **tqdm_kwargs) if _is_mapping(awaitables)
        else tqdm_asyncio.gather(*awaitables, return_exceptions=return_exceptions, **tqdm_kwargs) if tqdm
        else asyncio.gather(*awaitables, return_exceptions=return_exceptions)
    )
    
async def gather_mapping(mapping: Mapping[KT, Awaitable[VT]], return_exceptions: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Dict[KT, VT]:
    """
    Concurrently awaits a mapping of awaitable objects and returns a dictionary of results.

    This function is designed to await a mapping of awaitable objects, where each key-value pair represents a unique awaitable. It enables concurrent execution and gathers results into a dictionary.

    Args:
        mapping (Mapping[KT, Awaitable[VT]]): A dictionary-like object where keys are of type KT and values are awaitable objects of type VT.
        return_exceptions (bool, optional): If True, exceptions are returned as results instead of raising them. Defaults to False.
        tqdm (bool, optional): If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Returns:
        Dict[KT, VT]: A dictionary with keys corresponding to the keys of the input mapping and values containing the results of the corresponding awaitables.

    Example:
        The 'results' dictionary will contain the awaited results, where keys match the keys in the 'mapping' and values contain the results of the corresponding awaitables.
        ```
        mapping = {'task1': async_function1(), 'task2': async_function2(), 'task3': async_function3()}

        results = await gather_mapping(mapping)
        ```
    """
    results = {k: None for k in mapping.keys()}  # return data in same order
    async for k, v in as_completed_mapping(mapping, return_exceptions=return_exceptions, aiter=True, tqdm=tqdm, **tqdm_kwargs):
        results[k] = v
    return results

_is_mapping = lambda awaitables: len(awaitables) == 1 and isinstance(awaitables[0], Mapping)




import asyncio
from typing import AsyncIterable, AsyncIterator, Iterable, Iterator, TypeVar

T = TypeVar('T')

class ASyncIterable(AsyncIterable[T], Iterable[T]):
    """An abstract iterable implementation that can be used in both a `for` loop and an `async for` loop."""
    def __iter__(self) -> Iterator[T]:
        yield from ASyncIterator.wrap(self.__aiter__())
    @classmethod
    def wrap(self, aiterable: AsyncIterable[T]) -> "ASyncWrappedIterable[T]":
        return ASyncWrappedIterable(aiterable)

class ASyncIterator(AsyncIterator[T], Iterator[T]):
    """An abstract iterator implementation that can be used in both a `for` loop and an `async for` loop."""
    def __next__(self) -> T:
        try:
            return asyncio.get_event_loop().run_until_complete(self.__anext__())
        except StopAsyncIteration as e:
            raise StopIteration from e
    @classmethod
    def wrap(self, aiterator: AsyncIterator[T]) -> "ASyncWrappedIterator[T]":
        return ASyncWrappedIterator(aiterator)

class ASyncWrappedIterable(ASyncIterable[T]):
    def __init__(self, async_iterable: AsyncIterable[T]):
        self.__aiterable = async_iterable
    def __aiter__(self) -> AsyncIterable[T]:
        return self.__aiterable.__aiter__()

class ASyncWrappedIterator(ASyncIterator[T]):
    def __init__(self, async_iterator: AsyncIterator[T]):
        self.__aiterator = async_iterator
    async def __anext__(self) -> T:
        return await self.__aiterator.__anext__()




from a_sync.singleton import ASyncGenericSingleton

def test_flag_predefined():
    """We had a failure case where the subclass implementation assigned the flag value to the class and did not allow user to determine at init time"""
    class Test(ASyncGenericSingleton):
        sync=True
        def __init__(self):
            ...
    Test()
    class TestInherit(Test):
        ...
    TestInherit()

    class Test(ASyncGenericSingleton):
        sync=False
    Test()
    class TestInherit(Test):
        ...
    TestInherit()


# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio
import functools

from a_sync import exceptions, primitives
from a_sync._typing import *

# We keep this here for now so we don't break downstream deps. Eventually will be removed.
from a_sync.primitives import ThreadsafeSemaphore, DummySemaphore


@overload
async def apply_semaphore(  # type: ignore [misc]
    coro_fn: Literal[None],
    semaphore: SemaphoreSpec,
) -> AsyncDecorator[P, T]:...

@overload
async def apply_semaphore(
    coro_fn: SemaphoreSpec,
    semaphore: Literal[None],
) -> AsyncDecorator[P, T]:...

@overload
async def apply_semaphore(
    coro_fn: CoroFn[P, T],
    semaphore: SemaphoreSpec,
) -> CoroFn[P, T]:...
    
def apply_semaphore(
    coro_fn: Optional[Union[CoroFn[P, T], SemaphoreSpec]] = None,
    semaphore: SemaphoreSpec = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    # Parse Inputs
    if isinstance(coro_fn, (int, asyncio.Semaphore)):
        if semaphore is not None:
            raise ValueError("You can only pass in one arg.")
        semaphore = coro_fn
        coro_fn = None
        
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)
        
    # Create the semaphore if necessary
    if isinstance(semaphore, int):
        semaphore = primitives.ThreadsafeSemaphore(semaphore)
    elif not isinstance(semaphore, asyncio.Semaphore):
        raise TypeError(f"'semaphore' must either be an integer or a Semaphore object.")
    
    # Create and return the decorator
    if isinstance(semaphore, primitives.Semaphore):
        # NOTE: Our `Semaphore` primitive can be used as a decorator.
        #       While you can use it the `async with` way like any other semaphore and we could make this code section cleaner,
        #       applying it as a decorator adds some useful info to its debug logs so we do that here if we can.
        semaphore_decorator = semaphore
    
    else:
        def semaphore_decorator(coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
            @functools.wraps(coro_fn)
            async def semaphore_wrap(*args, **kwargs) -> T:
                async with semaphore:  # type: ignore [union-attr]
                    return await coro_fn(*args, **kwargs)
            return semaphore_wrap
        
    return semaphore_decorator if coro_fn is None else semaphore_decorator(coro_fn)


dummy_semaphore = primitives.DummySemaphore()




import asyncio
import functools
from concurrent.futures import Executor
from inspect import getfullargspec

from async_lru import _LRUCacheWrapper
from async_property.base import \
    AsyncPropertyDescriptor  # type: ignore [import]
from async_property.cached import \
    AsyncCachedPropertyDescriptor  # type: ignore [import]

from a_sync import _flags
from a_sync._typing import *
from a_sync.exceptions import (ASyncRuntimeError, KwargsUnsupportedError,
                               SyncModeInAsyncContextError)


def get_event_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError as e: # Necessary for use with multi-threaded applications.
        if not str(e).startswith("There is no current event loop in thread"):
            raise e
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop
    
def _validate_wrapped_fn(fn: Callable) -> None:
    """Ensures 'fn' is an appropriate function for wrapping with a_sync."""
    if isinstance(fn, (AsyncPropertyDescriptor, AsyncCachedPropertyDescriptor)):
        return # These are always valid
    if not callable(fn):
        raise TypeError(f'Input is not callable. Unable to decorate {fn}')
    if isinstance(fn, _LRUCacheWrapper):
        fn = fn.__wrapped__
    fn_args = getfullargspec(fn)[0]
    for flag in _flags.VIABLE_FLAGS:
        if flag in fn_args:
            raise RuntimeError(f"{fn} must not have any arguments with the following names: {_flags.VIABLE_FLAGS}")

def _await(awaitable: Awaitable[T]) -> T:
    try:
        return get_event_loop().run_until_complete(awaitable)
    except RuntimeError as e:
        if str(e) == "This event loop is already running":
            raise SyncModeInAsyncContextError from e
        raise e

def _asyncify(func: SyncFn[P, T], executor: Executor) -> CoroFn[P, T]:  # type: ignore [misc]
    @functools.wraps(func)
    async def _asyncify_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.futures.wrap_future(
            executor.submit(func, *args, **kwargs), 
            loop=get_event_loop(),
        )
    return _asyncify_wrap




import inspect
import logging
from contextlib import suppress
from functools import cached_property
from typing import Any, Dict, Optional

from a_sync import _flags, exceptions
from a_sync.abstract import ASyncABC


logger = logging.getLogger(__name__)

class ASyncGenericBase(ASyncABC):
    """
    Inherit from this class to a-syncify all of your bound methods.
    Allows for the use of a variety of flags out-of-box.
    You can choose which flag(s) work best for your subclass implementation.
    """
    @cached_property
    def __a_sync_flag_name__(self) -> str:
        logger.debug("checking a_sync flag for %s", self)
        try:
            flag = self.__get_a_sync_flag_name_from_signature()
        except exceptions.ASyncFlagException:
            # We can't get the flag name from the __init__ signature,
            # but maybe the implementation sets the flag somewhere else.
            # Let's check the instance's atributes
            logger.debug("unable to find flag name using `%s.__init__` signature, checking for flag attributes defined on %s", self.__class__.__name__, self)
            present_flags = [flag for flag in _flags.VIABLE_FLAGS if hasattr(self, flag)]
            if len(present_flags) == 0:
                raise exceptions.NoFlagsFound(self)
            if len(present_flags) > 1:
                raise exceptions.TooManyFlags(self, present_flags)
            flag = present_flags[0]
        if not isinstance(flag, str):
            raise exceptions.InvalidFlag(flag)
        return flag
        
    @property  # This property is not cached so you can 
    def __a_sync_flag_value__(self) -> bool:
        flag = self.__a_sync_flag_name__
        flag_value = getattr(self, flag)
        if not isinstance(flag_value, bool):
            raise exceptions.InvalidFlagValue(flag, flag_value)
        logger.debug('`%s.%s` is currently %s', self, flag, flag_value)
        return flag_value

    @classmethod  # type: ignore [misc]
    def __a_sync_default_mode__(cls) -> bool:
        try:
            flag = cls.__get_a_sync_flag_name_from_signature()
            flag_value = cls.__a_sync_flag_default_value_from_signature()
        except exceptions.NoFlagsFound:
            flag = cls.__get_a_sync_flag_name_from_class_def()
            flag_value = cls.__get_a_sync_flag_value_from_class_def(flag)
        sync = _flags.negate_if_necessary(flag, flag_value)  # type: ignore [arg-type]
        logger.debug("`%s.%s` indicates default mode is %ssynchronous", cls, flag, 'a' if sync is False else '')
        return sync
    
    @classmethod
    def __get_a_sync_flag_name_from_signature(cls) -> Optional[str]:
        logger.debug("Searching for flags defined on %s.__init__", cls)
        if cls.__name__ == "ASyncGenericBase":
            logger.debug("There are no flags defined on the base class, this is expected. Skipping.")
            return None
        parameters = inspect.signature(cls.__init__).parameters
        logger.debug("parameters: %s", parameters)
        return cls.__parse_flag_name_from_list(parameters)

    @classmethod
    def __get_a_sync_flag_name_from_class_def(cls) -> Optional[str]:
        logger.debug("Searching for flags defined on %s", cls)
        try:
            return cls.__parse_flag_name_from_list(cls.__dict__)
        except exceptions.NoFlagsFound:
            for base in cls.__bases__:
                with suppress(exceptions.NoFlagsFound):
                    return cls.__parse_flag_name_from_list(base.__dict__)
        raise exceptions.NoFlagsFound(cls, list(cls.__dict__.keys()))

    @classmethod  # type: ignore [misc]
    def __a_sync_flag_default_value_from_signature(cls) -> bool:
        logger.debug("checking `__init__` signature for default %s a_sync flag value", cls)
        signature = inspect.signature(cls.__init__)
        flag = cls.__get_a_sync_flag_name_from_signature()
        flag_value = signature.parameters[flag].default
        if flag_value is inspect._empty:  # type: ignore [attr-defined]
            raise NotImplementedError(
                "The implementation for 'cls' uses an arg to specify sync mode, instead of a kwarg. We are unable to proceed. I suppose we can extend the code to accept positional arg flags if necessary"
            )
        logger.debug('%s defines %s, default value %s', cls, flag, flag_value)
        return flag_value

    @classmethod
    def __get_a_sync_flag_value_from_class_def(cls, flag: str) -> Optional[bool]:
        for spec in [cls, *cls.__bases__]:
            flag_value = spec.__dict__.get(flag)
            if flag_value is not None:
                return flag_value

    @classmethod
    def __parse_flag_name_from_list(cls, items: Dict[str, Any]) -> Optional[str]:
        present_flags = [flag for flag in _flags.VIABLE_FLAGS if flag in items]
        if len(present_flags) == 0:
            logger.debug("There are too many flags defined on %s", cls)
            raise exceptions.NoFlagsFound(cls, items.keys())
        if len(present_flags) > 1:
            logger.debug("There are too many flags defined on %s", cls)
            raise exceptions.TooManyFlags(cls, present_flags)
        flag = present_flags[0]
        logger.debug("found flag %s", flag)
        return flag



import abc
import asyncio
from typing import Optional

from a_sync.primitives._loggable import _LoggerMixin


class _DebugDaemonMixin(_LoggerMixin, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def _debug_daemon(self, fut: asyncio.Future, fn, *args, **kwargs) -> None:
        ...    
    def _start_debug_daemon(self, *args, **kwargs) -> "asyncio.Future[None]":
        if self.debug_logs_enabled and asyncio.get_event_loop().is_running():
            return asyncio.create_task(self._debug_daemon(*args, **kwargs))
        # else we return a blank Future since we shouldn't or can't create the daemon
        return asyncio.get_event_loop().create_future()
    def _ensure_debug_daemon(self, *args, **kwargs) -> "asyncio.Future[None]":
        if not self.debug_logs_enabled:
            self._daemon = asyncio.get_event_loop().create_future()
        if not hasattr(self, '_daemon') or self._daemon is None:
            self._daemon = self._start_debug_daemon(*args, **kwargs)
            self._daemon.add_done_callback(self._stop_debug_daemon)
        return self._daemon
    def _stop_debug_daemon(self, t: Optional[asyncio.Task] = None) -> None:
        if t and t != self._daemon:
            raise ValueError(f"{t} is not {self._daemon}")
        self._daemon.cancel()
        self._daemon = None




import logging
import threading
from abc import ABCMeta
from typing import Any, Dict, Tuple

from a_sync import ENVIRONMENT_VARIABLES, _bound, modifiers
from a_sync.future import _ASyncFutureWrappedFn
from a_sync.modified import ASyncFunction, Modified
from a_sync.property import PropertyDescriptor

logger = logging.getLogger(__name__)

class ASyncMeta(ABCMeta):
    """Any class with metaclass ASyncMeta will have its functions wrapped with a_sync upon class instantiation."""
    def __new__(cls, new_class_name, bases, attrs):
        _update_logger(new_class_name)
        logger.debug(f"woah, you're defining a new ASync class `%s`! let's walk thru it together", new_class_name)
        logger.debug(f"first, I check whether you've defined any modifiers on `%s`", new_class_name)
        # NOTE: Open uesion: what do we do when a parent class and subclass define the same modifier differently?
        #       Currently the parent value is used for functions defined on the parent, 
        #       and the subclass value is used for functions defined on the subclass.
        class_defined_modifiers = modifiers.get_modifiers_from(attrs)
        logger.debug(f'found modifiers: %s', class_defined_modifiers)
        logger.debug("now I inspect the class definition to figure out which attributes need to be wrapped")
        for attr_name, attr_value in list(attrs.items()):
            if attr_name.startswith("_"):
                logger.debug(f"`%s.%s` starts with an underscore, skipping", new_class_name, attr_name)
                continue
            elif "__" in attr_name:
                logger.debug(f"`%s.%s` incluldes a double-underscore, skipping", new_class_name, attr_name)
                continue
            elif isinstance(attr_value, _ASyncFutureWrappedFn):
                logger.debug(f"`%s.%s` is a %s, skipping", new_class_name, attr_name, attr_value.__class__.__name__)
                continue
            logger.debug(f"inspecting `{new_class_name}.{attr_name}` of type {attr_value.__class__.__name__}")
            fn_modifiers = dict(class_defined_modifiers)
            # Special handling for functions decorated with a_sync decorators
            if isinstance(attr_value, Modified):
                logger.debug(f"`{new_class_name}.{attr_name}` is a `Modified` object, which means you decorated it with the a_sync decorator even though `{new_class_name}` is an ASync class")
                logger.debug(f"you probably did this so you could apply some modifiers to `{attr_name}` specifically")
                modified_modifiers = attr_value.modifiers._modifiers
                if modified_modifiers:
                    logger.debug(f"I found `{new_class_name}.{attr_name}` is modified with {modified_modifiers}")
                    fn_modifiers.update(modified_modifiers)
                else:
                    logger.debug(f"I did not find any modifiers")
                logger.debug(f"full modifier set for `{new_class_name}.{attr_name}`: {fn_modifiers}")
                if isinstance(attr_value, PropertyDescriptor):
                    # Wrap property
                    logger.debug(f"`{attr_name} is a property, now let's wrap it")
                    wrapped, hidden = _bound._wrap_property(attr_value, **fn_modifiers)
                    attrs[attr_name] = wrapped
                    logger.debug(f"`{attr_name}` is now `{wrapped}`")
                    logger.debug(f"since `{attr_name}` is a property, we will add a hidden dundermethod so you can still access it both sync and async")
                    attrs[attr_value.hidden_method_name] = hidden
                    logger.debug(f"`{new_class_name}.{attr_value.hidden_method_name}` is now {hidden}")
                elif isinstance(attr_value, ASyncFunction):
                    attrs[attr_name] = _bound._wrap_bound_method(attr_value, **fn_modifiers)
                else:
                    raise NotImplementedError(attr_name, attr_value)
                    
            elif callable(attr_value):
                # NOTE We will need to improve this logic if somebody needs to use it with classmethods or staticmethods.
                attrs[attr_name] = _bound._wrap_bound_method(attr_value, **fn_modifiers)
            else:
                logger.debug(f"`{new_class_name}.{attr_name}` is not callable, we will take no action with it")
        return super(ASyncMeta, cls).__new__(cls, new_class_name, bases, attrs)    


class ASyncSingletonMeta(ASyncMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]) -> None:
        cls.__instances: Dict[bool, object] = {}
        cls.__lock = threading.Lock()
        super().__init__(name, bases, namespace)

    def __call__(cls, *args: Any, **kwargs: Any):
        is_sync = cls.__a_sync_instance_will_be_sync__(args, kwargs)  # type: ignore [attr-defined]
        if is_sync not in cls.__instances:
            with cls.__lock:
                # Check again in case `__instance` was set while we were waiting for the lock.
                if is_sync not in cls.__instances:
                    cls.__instances[is_sync] = super().__call__(*args, **kwargs)
        return cls.__instances[is_sync]

def _update_logger(new_class_name: str) -> None:
    if ENVIRONMENT_VARIABLES.DEBUG_MODE or ENVIRONMENT_VARIABLES.DEBUG_CLASS_NAME == new_class_name:
        logger.addHandler(_debug_handler)
        logger.setLevel(logging.DEBUG)
        logger.info("debug mode activated")
    else:
        logger.removeHandler(_debug_handler)
        logger.setLevel(logging.INFO)

_debug_handler = logging.StreamHandler()




from a_sync.base import ASyncGenericBase
from a_sync._meta import ASyncSingletonMeta

class ASyncGenericSingleton(ASyncGenericBase, metaclass=ASyncSingletonMeta):
    """
    Subclass this class if you want Singleton-esque ASync objects.
    They work kind of like a typical Singleton would, but you will have two instances instead of one:
    - one sync instance
    - one async instance
    """
    pass
